/*
 *  channel.cpp - Base class for the different channel operations
 *
 *  Copyright (c) 2015 Martin Glueck <martin@mangari.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3 as
 *  published by the Free Software Foundation.
 */

#include "channel.h"
#include "app_4sense.h"

#include <sblib/eib/com_objects.h>

void Channel::init(void)
{
	uint8_t temp_var;

	gw_init_done = false;
    for (uint8_t n=0; n<2; n++)            // configure cyclic sending
    {
        messwerte[n]=0;
        lasttemp[n]=0;
        lastsend[n]=0;

        // zykl. Senden Messwerte Basis
        timerbase[n] = (userEeprom.getUInt8(THBASE_ZYKLSEND+((id*4+n)>>1)) >> (4 *((id*4+n)&0x01))) &0x0F;

        // Verhalten bei Busspannungswiederkehr, 2 Messwerte je Kanal mit gleichem Faktor laden
        temp_var = userEeprom.getUInt8(RE_MW_S_FAKT+((id*4+n)>>1));
        if(temp_var & 0x80)  // Verzögert senden eingeschaltet
        {
            timercnt[n] = temp_var;
        }
        else
        {
            // Gleich zyklisch senden Faktor laden
            timercnt[n] = userEeprom.getUInt8(TEMP_ZYKLSEND +(id*4+n));
        }
    }

    // zyk. Senden Basis Grenzwerte, 4x
    timerbasegw = (userEeprom.getUInt8(GW_ZYKL_BASE +2*id) >>4);
    // zyk. Senden Faktor Grenzwerte wird in grenzwerte() initialisiert nach erfolgtem GW Init!
    timercntgw = 0x00;   // Abschalten TODO ist das nötig??

    switch (stype)
    {
    case TYPE_TEMP:
    	numval = 1;
    	break;
    case TYPE_HUM:
    	numval = 2;
        break;
    case TYPE_VOC:
    	numval = 2;
    	break;
    default:
    	numval = 1;
    	break;
    }
}

/* get measurement values from sensor object and add offset */
void Channel::read(void)
{
   //sensor->read(); // sid?
   for (uint8_t i=0; i<numval; i++)
   {
       // lasttemp[i]  = messwerte[i];	         // store last value: skip, done in checkPeriodic()
       messwerte[i] = sensor->getValue(i);       // store current value
       messwerte[i] = messwerte[i] + (signed char) userEeprom.getUInt8(TEMP_OFFSET+ 2*id + i);	    // Add offset
   }
}

/*
* Send on exceeding or falling below threshold
*   check thresholds, write objects and send telegram
*/
void Channel::checkPeriodic(void)
{
	int threshold;              // threshold, temp var
    unsigned char action;       // action on threshold exceede
    unsigned char th_type;      // threshold type (temperature or humidity)
    unsigned char objno;        // Messwert im Array
    unsigned char help_obj;     // Hilfsvariable
    uint8_t i;
    uint8_t grenzwert_eingang;    // zu sendener Wert bei Über-Unterschreiten
    uint8_t gw_changed;           // threshold has changed
    uint8_t gw_init=0;            // Initialize threshold

    int mess_diff;              // if limit of difference is exceeded, send value
    int mess_change;            // difference between last and current measurement
    uint8_t lee;                // local copy of TEMP_MESSDIFF

    // Threshold is only available for one measurement
    // Check if thresholds need to be initialized and send after start
    if (!gw_init_done)
    {
        gw_init_done = true;
        gw_init = 1;
        switch ((userEeprom.getUInt8(RE_SEND_GW)>>(2*id)) & 0x03)
        {
           case 1:			// Send directly
        	   timercntgw = 0x80;
           	   break;
           case 2:          // delayed sending
        	   timercntgw = userEeprom.getUInt8(RE_GW_S_FAKT) | 0x80;
           	   break;
           default:         // nicht senden -> cyclic sending of factor Grenzwerte
        	   timercntgw = userEeprom.getUInt8(GW_ZYKL_FAKT+id);
               break;
        }
    }

    // Threshold assignment: humidity (bit 1), temperature would be (bit 0) (!?)
    th_type = (userEeprom.getUInt8(TH_TYPE+id) >> 1);

    for (i=0; i<3; i++)			// do for all 3 thresholds
    {
        grenzwert_eingang = 0;  // clear
        gw_changed = 0;         // clear

        help_obj = id * 3 + i;                 // threshold objects 0..11
        objno = 2 * id + (th_type & 0x01);     // is current threshold for temperature or humidity?

        action = (userEeprom.getUInt8(GW_REAKTION + (2*id + (i>>1))) >> (4 *(i & 0x01))) & 0x0F;
        threshold = (int)((userEeprom.getUInt8(GW_SCHWELLE1L+(help_obj*2))<<8) | userEeprom.getUInt8(GW_SCHWELLE1H+(help_obj*2)))*10; // 16 bit value * 10, tbc

        if ((lasttemp[objno]<threshold || gw_init) && messwerte[objno]>threshold) {  // threshold exceeded
            if (action & 0x08)    // action on exceed is active
            {
                grenzwert_eingang = ((action >> 2) & 0x01);
                gw_changed = 1;
            }
        }

        if ((lasttemp[objno]>threshold || gw_init) && messwerte[objno]<threshold) {  // theshold fallen below
            if (action & 0x02)    // action on fall below is active
            {
                grenzwert_eingang = (action & 0x01);
                gw_changed = 1;
            }
        }

        if (gw_changed)
        {
            if (!gw_init)
            	objectWrite (COMOBJ_THRESHOLD0 + help_obj, grenzwert_eingang);  // Neuen Grenzwert dem Eingangsobjekt zuordnen
            else
            	objectSetValue (COMOBJ_THRESHOLD0 + help_obj, 0);     // threshold init just done, do NOT send during initialization
            lasttemp[objno]=messwerte[objno];        // store last value !?
        }
        th_type = th_type>>2;           // next assignment
    }

    // Send if difference of measurement value is exceeded
    // check if value valid !?
    for (i=0; i<numval; i++)    // do for all measurement values
    {
        lee = userEeprom.getUInt8(TEMP_MESSDIFF+id*2+i);   // get configuration from eeprom
        if (lee & 0x80)                    // if enabled check if limit is exceeded
        {
            mess_diff = (lee & 0x7F) * 10;
            mess_change = lastsend[i] - messwerte[i];
            if( abs(mess_change) > mess_diff )  // TODO must be >= if VD is doesn't allow 0 anymore ?
            {
            	objectWrite (COMOBJ_VALUE0+i, messwerte[i]);
            }
        }
    }
}

